/*******************************************************************************
                     Computación Paralela 2012 - FaMAF
                      Universidad Nacional de Córdoba

Ejercicio 1 - Normalización / contrast stretching
=================================================
El siguiente código realiza contrast stretching o normalización de una imagen de
escala de grises, modificando los valores de cada pixel de modo que los que
tenían valor máximo pasen a ser pixels blancos, los que tenían el valor mínimo
pasen a ser negros, y el resto se ajusta de manera lineal entre los valores
máximos y mínimos.

a) Escriba y llame el kernel que realiza la normalización. El kernel debería
   llamar a la función de device "normalize", que realiza la normalización por
   pixel.
b) Las reducciones para obtener máximos y mínimos son lentas. Implemente kernels
   con algoritmos de reducción O(n log n) que usen shared memory. Es probable que
   necesite cambiar el tamaño de dev_min y dev_max para almacenar valores
   temporales.

********************************************************************************/

#include <cuda.h>
#include <cutil.h>
#include <stdio.h>
#include <stdlib.h>

#include "pixelformats.h"
#include "sdlstuff.h"

#define BLOCK_X 16
#define BLOCK_Y 16


// helper
uint divceil(uint a, uint b) {
    return (a + b - 1) / b;
}


// normaliza un punto
__device__ grayscale8 normalize(grayscale8 value, grayscale8 old_min, grayscale8 old_max) {
    grayscale8 new_min = 0;
    grayscale8 new_max = 255;
    return (grayscale8) ((float) (value - old_min) * (new_max - new_min) / (old_max - old_min)) + new_min;
}


// FALTA: Lento!
__global__ void get_min(const grayscale8 * image, size_t width, size_t height, uint * minvalue) {
    uint x = blockIdx.x * blockDim.x + threadIdx.x;
    uint y = blockIdx.y * blockDim.y + threadIdx.y;
    size_t idx = x + y * width;
    if (x < width && y < height) {
        atomicMin(minvalue, image[idx]);
    }
}


// FALTA: Lento!
__global__ void get_max(const grayscale8 * image, size_t width, size_t height, uint * maxvalue) {
    uint x = blockIdx.x * blockDim.x + threadIdx.x;
    uint y = blockIdx.y * blockDim.y + threadIdx.y;
    size_t idx = x + y * width;
    if (x < width && y < height) {
        atomicMax(maxvalue, image[idx]);
    }
}


// FALTA: kernel para normalizar!


int main(int argc, char ** argv) {

    if (argc != 2) {
        fprintf(stderr, "Error: normalize [nombre de imagen]\n");
        exit(1);
    }

    // cargar imagen
    grayscale8 * host_orig;
    size_t width;
    size_t height;
    host_orig = sdls_loadimage_grayscale8(argv[1], &width, &height);
    if (host_orig == 0) {
        fprintf(stderr, "Error: Problemas abriendo la imagen\n");
        exit(1);
    }

    // copiar al device
    grayscale8 * dev_orig;
    CUDA_SAFE_CALL(cudaMalloc(&dev_orig, width * height * sizeof(grayscale8)));
    CUDA_SAFE_CALL(cudaMemcpy(dev_orig, host_orig, width * height * sizeof(grayscale8), cudaMemcpyDefault));

    // lugar para la imagen normalizada
    grayscale8 * dev_norm;
    CUDA_SAFE_CALL(cudaMalloc(&dev_norm, width * height * sizeof(grayscale8)));

    // pedir lugar para mínimo e inicializar al máximo valor
    uint * dev_min;
    CUDA_SAFE_CALL(cudaMalloc(&dev_min, sizeof(uint)));
    CUDA_SAFE_CALL(cudaMemset(dev_min, 0xff, sizeof(uint)));

    // pedir lugar para máximo e inicializar al mínimo valor
    uint * dev_max;
    CUDA_SAFE_CALL(cudaMalloc(&dev_max, sizeof(uint)));
    CUDA_SAFE_CALL(cudaMemset(dev_max, 0, sizeof(uint)));

    // buscar mínimo y máximo para poder normalizar
    dim3 block(BLOCK_X, BLOCK_Y);
    dim3 grid(divceil(width, block.x), divceil(height, block.y));
    get_min<<<grid, block>>>(dev_orig, width, height, dev_min);
    get_max<<<grid, block>>>(dev_orig, width, height, dev_max);
    CUT_CHECK_ERROR("Error en reducción: ");


    // FALTA: normalizar!


    // copiar imagen normalizada al host
    grayscale8 * host_norm = (grayscale8 *) malloc(width * height * sizeof(grayscale8));
    CUDA_SAFE_CALL(cudaMemcpy(host_norm, dev_norm, width * height * sizeof(grayscale8), cudaMemcpyDefault));

    // inicializar sdl
    sdls_init(2 * width + 1, height);

    // dibujar en pantalla
    sdls_drawrectangle_grayscale8(0, 0, width, height, host_orig);
    sdls_drawrectangle_grayscale8(width + 1, 0, width, height, host_norm);

    printf("<ENTER> para salir\n");
    getchar();

    // limpieza
    free(host_orig);
    free(host_norm);
    CUDA_SAFE_CALL(cudaFree(dev_orig));
    CUDA_SAFE_CALL(cudaFree(dev_norm));
    CUDA_SAFE_CALL(cudaFree(dev_min));
    CUDA_SAFE_CALL(cudaFree(dev_max));

    return 0;
}
